# -*- coding: utf-8 -*-
#    I, for one, welcome our new robot overloards
#    Copyright (C) 2008 David Eyk, David Thulson, Collin Sanford, David Howery
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""actors -- actors moving on the screen

$Author: csanford $
$Rev: 1034 $
$Date: 2008-04-05 18:25:10 -0500 (Sat, 05 Apr 2008) $
"""

__author__ = "$Author: csanford $"
__revision__ = "$Rev: 1034 $"
__date__ = "$Date: 2008-04-05 18:25:10 -0500 (Sat, 05 Apr 2008) $"

import rabbyt
from pyglet import image
from pyglet.window import key as keys

import resources
import rsprite
from entity import Entity, EntityCategories, CollidableCategories
from sound import sound
import terrain

from cocos import actions

from math import cos, sin, radians
import random

from projectx import ai

def raise_(msg):
    raise Exception(msg)

class Steerable(object):
    def __init__(self, rotation_rate=270, friction_rate=-4.5,
                 acceleration_rate=45, backing_rate=-25):
        self.rotation_rate = rotation_rate
        self.friction_rate = friction_rate
        self.acceleration_rate = acceleration_rate
        self.backing_rate = backing_rate

        self.keys_pressed = set()
        self.accelerating = False
        self.turning_right = False
        self.turning_left = False
        self.backing = False

        self.xy = [0,0]
        self.velocity = [0,0]

    def on_key_press(self, symbol, modifiers):
        self.keys_pressed.add (symbol)
        return self.handle_keys()

    def on_key_release(self, symbol, modifiers):
        try:
            self.keys_pressed.remove(symbol)
        except KeyError:
            pass
        return self.handle_keys()

    def handle_keys(self):
        pressed = self.keys_pressed
        self.accelerating = keys.UP in pressed
        self.turning_right = keys.RIGHT in pressed
        self.turning_left = keys.LEFT in pressed
        self.backing = keys.DOWN in pressed
        return True

        self.accelerating = False
        self.backing = False
        self.turning_right = False
        self.turning_left = False

    def updateMovement(self, dt):
        rot_delta = self.rotation_rate*dt
        if self.turning_right:
            self.rot -= rot_delta
        if self.turning_left:
            self.rot += rot_delta
            
        x, y = self.getVectorComponents()
        
        velocity = self.velocity

        # Friction is just an acceleration in the opposite direction than the
        #   current velocity
        f_delta = self.friction_rate*dt
        a_x = velocity[0]*f_delta
        a_y = velocity[1]*f_delta
        
        if self.accelerating:
            a_delta = self.acceleration_rate*dt
            a_x += x*a_delta
            a_y += y*a_delta
        elif self.backing:
            b_delta = self.backing_rate*dt
            a_x += x*b_delta
            a_y += y*b_delta

        velocity[0] += a_x
        velocity[1] += a_y

        self.x += velocity[0]
        self.y += velocity[1]


class Missile(Entity, Steerable):
    """A missile fired from the robots"""
    missile_img = resources.loadImage('missile.png')
    missile_smoke = resources.loadImage('star.png')
    missile_explode = resources.loadImage('explode.png')
    missile_explode_sound = sound.loadSound(resources.getFilePath('explode.wav'))
    
    def __init__(self, location, rotation, velocity=(0,0)):
        Entity.__init__(self, self.missile_img.texture,
                        EntityCategories.weapon)
        Steerable.__init__(self, acceleration_rate=60, backing_rate=0,
                           friction_rate=-2.5)
        self.rot = rotation
        self.xy = location
        self.z = 5
        
        self.alive = True
        
        self.accelerating = True
        
        self.mass = 1
        self.particles = []
        self.health = 0.1
        
    def step(self, dt):
        if not self.alive:
            return
        self.create_smoke_trail()
        self.updateMovement(dt)
        Entity.step(self,dt)
        
    def kill(self):
        Entity.kill(self)
        self._world.remove(self)
        ai.Root.deregisterObject(self, self.category)
        
    def damage(self, amt):
        if amt >= self.health:
            self.health = 0
            self.create_explosion()
            
    def create_explosion(self):
        s = rsprite.RASprite(self.missile_explode.texture)
        s.scale = 2.0
        
        lifetime = 2
        
        self.missile_explode_sound.play()
        
        r = random.random

        for i in xrange(5): 
            s = rsprite.RASprite(self.missile_explode.texture)
            s.scale = 1.0
            
            x, y = self.convert_offset((0,0))
            
            s.x = rabbyt.lerp(x+r()*10-5, x+r()*60-30, dt=lifetime)
            s.y = rabbyt.lerp(y+r()*10-5, y+r()*60-30, dt=lifetime)
            
            s.rot = rabbyt.lerp(0, 90*random.choice(range(-2,3)), dt=6)
            s.scale = rabbyt.lerp(1, random.choice((2,3,4)), dt=lifetime)
            
            now = rabbyt.get_time()
            self.particles.append(s)
            
        rabbyt.scheduler.add(rabbyt.get_time() + lifetime,
                             lambda:self.kill())
    
    def create_smoke_trail(self):
        s = rsprite.RASprite(self.missile_smoke.texture)
        s.scale = 4.0
        
        lifetime = 2
        
        x, y = self.convert_offset((0,0))
        
        r = random.random
        s.x = rabbyt.lerp(x+r()*10-5, x+r()*60-30, dt=lifetime)
        s.y = rabbyt.lerp(y+r()*10-5, y+r()*60-30, dt=lifetime)
        
        s.rot = rabbyt.lerp(0, 90*random.choice(range(-2,3)), dt=6)
        s.scale = rabbyt.lerp(1, random.choice(range(2,8)), dt=lifetime)
        
        now = rabbyt.get_time()
        s.rgba = .9, .9, .9, rabbyt.lerp(.8, 0, now+lifetime/2, now+lifetime)
        
        self.particles.append(s)
        rabbyt.scheduler.add(rabbyt.get_time() + lifetime,
                             lambda:self.particles.remove(s))
        
    def render(self):
        if self.alive:
            Entity.render(self)
        rabbyt.render_unsorted(self.particles)
        
    def collideWith(self, obj):
        obj.damage(10)
        self.accelerating = False
        self.velocity = [0,0]
        self.alive = False
        
        entities = self._world.queryEntitiesNear(self.x, self.y, 50, category_set=CollidableCategories)
        for i in entities:
            if i is not self and i is not obj and type(i) is not Missile:
                i.damage(5)

class Dude(Entity, Steerable):
    """A dude who runs around on the screen.
    """
    dude_standing = resources.loadImage('dude/dude-standing.png')
    dude_walking_1 = resources.loadImage('dude/dude-walking_1.png')
    dude_walking_2 = resources.loadImage('dude/dude-walking_2.png')
    dude_running_1 = resources.loadImage('dude/dude-running_1.png')
    dude_running_2 = resources.loadImage('dude/dude-running_2.png')
    dude_shadow = resources.loadImage('dude/dude-shadow.png')
    dude_squished = resources.loadImage('dude/dude-squished.png')
    dude_squished_sound = sound.loadSound(resources.getFilePath('sfx/squish.wav'))
    dude_arg_sound = sound.loadSound(resources.getFilePath('sfx/aaagh.wav'))    
    dude_help_sound = sound.loadSound(resources.getFilePath('sfx/help-02.wav'))
    dude_no_sound = sound.loadSound(resources.getFilePath('sfx/no-01.wav'))
    dude_help_me_sound = sound.loadSound(resources.getFilePath('sfx/help-me-01.wav'))

    time_between_ai_updates = 1/4
    
    def __init__(self, use_ai=True):
        Entity.__init__(self, self.dude_standing.texture,
                               EntityCategories.other)
        Steerable.__init__(self, acceleration_rate=20,
                           backing_rate=-20)

        self.animation = self.updateAnimation()

        self.shadow = rsprite.RASprite(self.dude_shadow.texture)
        self.shadow.rot = self.attrgetter('rot')
        self.shadow.x = self.attrgetter('x') - 4
        self.shadow.y = self.attrgetter('y') - 5

        self.last_collision = None
        self.accelerating = False
        self.backing = False
        self.turning_right = False
        self.turning_left = False
        
        self.alive = True

        self.xy = [0,0]
        self.velocity = [0,0]
        
        self.mass = 1

        self._use_ai = use_ai
        if use_ai:
            self._time_to_next_ai = random.random() * self.time_between_ai_updates
            self.ai = ai.factory.buildDudeAI(self)
            self.step = self.stepWithAI
    
    def _random_help(self):
        if random.random() <= 0.0001:
            s = random.choice((self.dude_help_sound,
                               self.dude_no_sound,
                               self.dude_help_me_sound))
            s.play()

        self.animation = self.updateAnimation()

    def updateAnimation(self):
        yield None
        alternate = True

        run_1 = self.dude_running_1.texture
        run_2 = self.dude_running_2.texture
        stand = self.dude_standing.texture

        while True:
            if self.accelerating:
                if alternate:
                    self.texture = run_1
                    alternate = False
                else:
                    self.texture = run_2
                    alternate = True
            else:
                self.texture = stand
            yield None
            

    def step(self, dt):
        if not self.alive:
            return
        self.updateMovement(dt)
        self.animation.next()
        self._random_help()
        Entity.step(self, dt)
        
    def stepWithAI(self, dt):
        if not self.alive:
            return
        self.updateMovement(dt)
        self.animation.next()
        self._random_help()
        self.ai.step(dt)
        Entity.step(self, dt)

        #if self._use_ai:
        #    self._time_to_next_ai -= dt
        #    if self._time_to_next_ai <= 0:
        #        self._time_to_next_ai += self.time_between_ai_updates
        #        self.ai.step(dt)

        Entity.step(self, dt)

    def render(self):
        self.shadow.render()
        Entity.render(self)
    
    def damage(self, amt):
        if amt > 1:
            self.health = 0
            self.kill()
            self.dude_squished_sound.play()
            if random.choice((False,False,True)):
                self.dude_arg_sound.play()
            self.texture = self.dude_squished.texture

    def collideWith(self, obj):
        if type(obj) == Missile:
            obj.damage(0.1) # Just enough for an explosion
        self.last_collision = obj.xy
        
    def kill(self):
        Entity.kill(self)
        self._world.change_points(-5, message="Don't let dudes die!")
        ai.Root.deregisterObject(self, self.category)

class Car(Entity, Steerable):
    """Reimplementation of the Rabbyt driving example.
    """
    car_img = image.load(resources.getFilePath('car.png'))

    star_img = image.load(resources.getFilePath("star.png"))
    shadow_img = image.load(resources.getFilePath('carshadow.png'))

    def __init__(self):
        Entity.__init__(self, self.car_img.texture,
                               EntityCategories.player)
        Steerable.__init__(self)
        
        self.mass = 2

        self.shadow = rsprite.RASprite(self.shadow_img.texture)
        self.shadow.alpha = .5

        self.dust_particles = set()
    
        self.shadow.rot = self.attrgetter('rot')
        self.shadow.x = self.attrgetter('x') - 4
        self.shadow.y = self.attrgetter('y') - 5

        self.dust_r = (-15, 10)
        self.dust_l = (-15,-10)

    def step(self, dt):
        if self.accelerating:
            self.create_dust_particle(self.dust_r)
            self.create_dust_particle(self.dust_l)
            
        self.updateMovement(dt)

        Entity.step(self, dt)

    def create_dust_particle(self, offset):
        s = rsprite.RASprite(self.star_img.texture)
        s.scale = 2.0

        lifetime = 4

        x, y = self.convert_offset(offset)

        r = random.random  # (shortcut)
        s.x = rabbyt.lerp(x+r()*10-5, x+r()*60-30, dt=lifetime)
        s.y = rabbyt.lerp(y+r()*10-5, y+r()*60-30, dt=lifetime)

        s.rot = rabbyt.lerp(0, 90*random.choice(range(-2,3)), dt=6)

        s.scale = rabbyt.lerp(1, 4, dt=lifetime)

        now = rabbyt.get_time()
        s.rgba = .7, .5, 0, rabbyt.lerp(.1, 0, now+lifetime/2, now+lifetime)

        self.dust_particles.add(s)
        rabbyt.scheduler.add(rabbyt.get_time()+lifetime,
                lambda:self.dust_particles.remove(s))

    def render(self):
        self.shadow.render()
        Entity.render(self)
        rabbyt.render_unsorted(self.dust_particles)

class Robot(Entity, Steerable):
    """A robot.
    """
    robot_img = image.load(resources.getFilePath('robot01/robot01.png'))
    enemy_img = image.load(resources.getFilePath('robot01/robot01-enemy.png'))
    shadow_img = image.load(resources.getFilePath('robot01/robot01-shadow.png'))
    flame_img = image.load(resources.getFilePath('robot01/robot01-flame.png'))
    flame_shadow_img = image.load(resources.getFilePath('robot01/robot01-flame-shadow.png'))
    star_img = image.load(resources.getFilePath("star.png"))
    fire_sound = sound.loadSound(resources.getFilePath("fire.wav"))

    damage1 = resources.loadImage('robot01/robot01-damage01.png')
    damage2 = resources.loadImage('robot01/robot01-damage02.png')
    damage3 = resources.loadImage('robot01/robot01-damage03.png')

    def __init__(self, enemy=False):
        if enemy:
            img = self.enemy_img
            category = EntityCategories.AI
            self.use_ai = True
        else:
            img = self.robot_img
            category = EntityCategories.player
            self.use_ai = False
        Entity.__init__(self, img.texture,
                               category)
        Steerable.__init__(self)

        self.health = 100
        self.overlay = None

        self.mass = 10

        self.shadow = rsprite.RASprite(self.shadow_img.texture)
        self.shadow.alpha = .5

        self.dust_particles = set()
    
        self.shadow.rot = self.attrgetter('rot')
        self.shadow.x = self.attrgetter('x') - 4
        self.shadow.y = self.attrgetter('y') - 5

        self.dust_r = (-15, 10)
        self.dust_l = (-15,-10)
        
        self.bounding_radius *= .8
        
        self.shoot_r = (40, 20)
        self.shoot_l = (40, -20)
        self.gun_alternator = False
        self.firing = False
        self._reload_time = 0

        if self.use_ai:
            self.ai = ai.factory.buildEnemyAI(self)
            self.step = self.stepWithAI

        self.last_collision = None

    
    def reloading(self, dt):
        if self._reload_time <= 0:
            return False
        self._reload_time -= dt
        return True
    
    def _fire(self):
        self.firing = False

        ga = self.gun_alternator
        if ga:
            xy = self.convert_offset(self.shoot_r)
        else:
            xy = self.convert_offset(self.shoot_l)
        m = Missile(xy, self.rot, (self.velocity[0]*1.5, self.velocity[1]*1.5))
        self._world.add(m)
        self._reload_time = 0.5

        self.gun_alternator = not ga
        self.fire_sound.play()
        m.step(0.2)

    def step(self, dt):
        if not self.alive:
            return

        if self.accelerating:
            self.create_dust_particle(self.dust_r)
            self.create_dust_particle(self.dust_l)
            
        if not self.reloading(dt) and self.firing:
            self._fire()
            
        self.updateMovement(dt)
        
        Entity.step(self, dt)
        
    def stepWithAI(self, dt):
        if not self.alive:
            return

        if self.accelerating:
            self.create_dust_particle(self.dust_r)
            self.create_dust_particle(self.dust_l)
            
        if not self.reloading(dt) and self.firing:
            self._fire()

        self.updateMovement(dt)

        self.ai.step(dt)
        Entity.step(self, dt)


    def handle_keys(self):
        self.firing = keys.SPACE in self.keys_pressed
        return Steerable.handle_keys(self)
    
    def create_dust_particle(self, offset):
        s = rsprite.RASprite(self.star_img.texture)
        s.scale = 2.0

        lifetime = 4

        x, y = self.convert_offset(offset)

        r = random.random  # (shortcut)
        s.x = rabbyt.lerp(x+r()*10-5, x+r()*60-30, dt=lifetime)
        s.y = rabbyt.lerp(y+r()*10-5, y+r()*60-30, dt=lifetime)

        s.rot = rabbyt.lerp(0, 90*random.choice(range(-2,3)), dt=6)

        s.scale = rabbyt.lerp(1, 4, dt=lifetime)

        now = rabbyt.get_time()
        s.rgba = .7, .5, 0, rabbyt.lerp(.1, 0, now+lifetime/2, now+lifetime)

        self.dust_particles.add(s)
        rabbyt.scheduler.add(rabbyt.get_time()+lifetime,
                lambda:self.dust_particles.remove(s))

    def render(self):
        self.shadow.render()
        Entity.render(self)
        overlay = self.overlay
        if overlay:
            overlay.render()
        rabbyt.render_unsorted(self.dust_particles)

    
    def collideWith(self, obj):
        if type(obj) is not terrain.Tile:
            DAMAGE_RATIO = 0.2
            s = self.speed
            obj.damage((s * DAMAGE_RATIO)+0.1)
        self.last_collision = obj.xy

    def addOverlayImg(self, overlay):
        s = rsprite.RASprite(overlay.texture)
        s.rot = self.attrgetter('rot')
        s.xy = self.attrgetter('xy')
        self.overlay = s

    def damage(self, amt):
        self.health -= amt
        health = self.health
        if self.health <= 0:
            self.health = 0
            self.kill()
            self.addOverlayImg(self.damage3)
        elif health <= 30:
            self.addOverlayImg(self.damage2)
        elif health <= 70:
            self.addOverlayImg(self.damage1)
            
    def kill(self):
        self.alive = False
        self.shadow.texture = self.damage3.texture
        self.z -= 0.5
        if self.use_ai:
            self._world.change_points((random.random()+ 1) * 10250, 
                                      message="Way to kill the enemy!!!")
        ai.Root.deregisterObject(self, self.category)
