# -*- coding: utf-8 -*-
#    I, for one, welcome our new robot overloards
#    Copyright (C) 2008 David Eyk, David Thulson, Collin Sanford, David Howery
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""ai -- artificial intelligence for actors, using Behavior Trees.

http://aigamedev.com/hierarchical-logic

$Author: eykd $
$Rev: 871 $
$Date: 2008-04-02 19:38:18 -0500 (Wed, 02 Apr 2008) $
"""

__author__ = "$Author: eykd $"
__revision__ = "$Rev: 871 $"
__date__ = "$Date: 2008-04-02 19:38:18 -0500 (Wed, 02 Apr 2008) $"

import sys
import traceback
import weakref
import operator
import random
choice = random.choice

from math import atan2, radians, degrees, pi
pi_2 = pi*2
pi_1_2 = pi/2.0

import rabbyt
from rabbyt import collisions
import enum

import world
from entity import EntityCategories

from cellulose import InputCellDescriptor, ComputedCellDescriptor

RESULT = enum.Enum('FAIL', 'SUCCEED', 'CONTINUE')

class _BaseBTNode(object):
    """Base Behavior Tree Node.
    """
    def __init__(self, **kwargs):
        super(_BaseBTNode, self).__init__()
        self.parent = None

        self.priority = kwargs.get('priority', 1)
        self.probability = kwargs.get('probability', 1)

        self.RESULT_TABLE = {RESULT.SUCCEED: self.ifChildSucceeds,
                             RESULT.FAIL: self.ifChildFails,
                             RESULT.CONTINUE: self.ifChildContinues,}

    @ComputedCellDescriptor
    def target(self):
        return self.parent.target

    @ComputedCellDescriptor
    def memory(self):
        return self.parent.memory

    @ComputedCellDescriptor
    def blackboard(self):
        return self.parent.blackboard

    def report(self, action_msg):
        print self.target.__class__.__name__, action_msg

    def isRoot(self):
        return self.parent is None

    def step(self, dt):
        return RESULT.SUCCEED

    def stepTimes(self, n, dt):
        dt = dt/float(n)
        results = tuple((self.step(dt) for x in xrange(n)))
        return results

    def ifChildSucceeds(self, child):
        return RESULT.SUCCEED

    def ifChildFails(self, child):
        return RESULT.FAIL

    def ifChildContinues(self, child):
        return RESULT.CONTINUE

    def step(self, dt):
        return RESULT.SUCCEED


class _BTNode(_BaseBTNode):
    """Behavior Tree Node.
    """
    def __init__(self, *children, **kwargs):
        super(_BTNode, self).__init__(**kwargs)
        self.children = []

        self.addChildren(*children)

    def addChildren(self, *children):
        [setattr(c, 'parent', self) for c in children]
        self.children.extend(children)

    def isLeaf(self):
        return not bool(self.children)

class _BTLeaf(_BTNode):
    def __init__(self, **kwargs):
        super(_BTLeaf, self).__init__(**kwargs)


class _BTDecorator(_BaseBTNode):
    """Special purpose Behavior Tree node that can have only one child.

    Defaults to identity behavior.
    """
    def __init__(self, child, **kwargs):
        super(_BTDecorator, self).__init__(**kwargs)
        self.child = None

        self.setChild(child)

    def setChild(self, child):
        """Set the decorator's child.
        """
        self.child = child
        self.child.parent = self

    def isLeaf(self):
        """Decorators can't be leaves.
        """
        return False

    def step(self, dt):
        c = self.child
        c_result = c.step(dt)
        return self.RESULT_TABLE[c_result](c)

class Identity(_BTDecorator):
    """Transparent decorator.  Runs its child and passes on the result.
    """
    pass


class Root(_BTDecorator):
    """Root behavior tree node.  

    Repeat child forever.  Pretend everything's okay, even if it's
    not. Report exceptions.
    """
    target = InputCellDescriptor()
    memory = InputCellDescriptor()
    blackboard = {}

    def __init__(self, target, child, initial_delay = 0.0):
        self.target = target
        self.memory = {}
        self.initial_delay = initial_delay
        self.counter = 0.0
        super(Root, self).__init__(child)

        if initial_delay > 0.0:
            self._step = self.step
            self.step = self._delayStep

    @classmethod
    def registerObject(cls, obj, category):
        c_list = cls.blackboard.setdefault(category, [])
        c_list.append(obj)
        cls.blackboard.setdefault('all_objects', []).append(obj)

    @classmethod
    def deregisterObject(cls, obj, category):
        c_list = cls.blackboard.setdefault(category, [])
        try:
            c_list.remove(obj)
        except ValueError:
            pass
        a_list = cls.blackboard.setdefault('all_objects', [])
        try:
            a_list.remove(obj)
        except ValueError:
            pass

    def _delayStep(self, dt):
        self.counter += dt
        if dt >= self.initial_delay:
            self.step = self._step
            self.step(dt)
        return RESULT.CONTINUE

    def step(self, dt):
        """Step through children.
        """
        try:
            self.child.step(dt)
        except Exception, e:
            print "Exception on", self.target
            traceback.print_exc()
        return RESULT.CONTINUE

class Sequence(_BTNode):
    """Sequence behavior.
    """
    def __init__(self, *children, **kwargs):
        super(Sequence, self).__init__(*children, **kwargs)
        self.cursor = 0

    def reset(self):
        self.cursor = 0

    def ifChildSucceeds(self, child):
        self.cursor += 1
        if self.cursor >= len(self.children):
            self.reset()
            result = RESULT.SUCCEED
        else:
            result = RESULT.CONTINUE
        return result

    def ifChildFails(self, child):
        self.reset()
        return RESULT.FAIL

    def step(self, dt):
        """Step through children in sequence.
        """
        try:
            c = self.children[self.cursor]
            c_result = c.step(dt)
        except Exception, e:
            self.reset()
            raise e
            
        return self.RESULT_TABLE[c_result](c)


_getProbability = operator.attrgetter('probability')
def pickByProbability(candidates):
    picklist = []
    for c in candidates:
        for x in xrange(_getProbability(c)):
            picklist.append(c)
    return choice(picklist)

_getPriority = operator.attrgetter('priority')
def pickByPriority(candidates):
    return sorted(candidates, key=_getPriority)[-1]

class Selector(_BTNode):
    """Selector behavior.

    Pick a child using the given picking strategy and run it.

    
    """
    def __init__(self, *children, **kwargs):
        super(Selector, self).__init__(*children, **kwargs)
        self.picker = kwargs.get('picker', pickByProbability)
        self.continue_child = False
        self.last_pick = None
    
    def ifChildContinues(self, child):
        self.continue_child = True
        return RESULT.CONTINUE

    def step(self, dt):
        if self.continue_child:
            self.continue_child = False
        else:
            self.last_pick = self.picker(self.children)
        
        c = self.last_pick
        c_result = c.step(dt)

        return self.RESULT_TABLE[c_result](c)


class RepeatUntilFail(_BTDecorator):
    """Repeat the decorated behavior unti it fails.
    """
    def ifChildSucceeds(self, child):
        return RESULT.CONTINUE

    def ifChildFails(self, child):
        return RESULT.FAIL


class RepeatUntilSucceed(_BTDecorator):
    """Repeat the decorated behavior unti it succeeds.
    """
    def ifChildSucceeds(self, child):
        return RESULT.SUCCEED

    def ifChildFails(self, child):
        return RESULT.CONTINUE

    def step(self, dt):
        c = self.child
        c_result = c.step(dt)
        return self.RESULT_TABLE[c_result](c)


FAILURE_POLICY = enum.Enum('ALL', 'ONE')
COMPLETION_POLICY = enum.Enum('ALL', 'ONE')

class Parallel(_BTNode):
    """Parallel behaviors.

    Run all children in parallel (arbitrary sequential order).
    """
    def __init__(self, *children, **kwargs):
        super(Parallel, self).__init__(*children, **kwargs)

        self.failure_policy = kwargs.get('failure_policy',
                                         FAILURE_POLICY.ONE)
        self.completion_policy = kwargs.get('completion_policy',
                                         COMPLETION_POLICY.ALL)

        self.failed = 0
        self.completed = 0

        self.completed_children = []

        self.POLICY = {FAILURE_POLICY.ONE: self._policyOneFails,
                       FAILURE_POLICY.ALL: self._policyAllFail,
                       COMPLETION_POLICY.ONE: self._policyOneCompletes,
                       COMPLETION_POLICY.ALL: self._policyAllComplete}

    def reset(self):
        self.children.extend(self.completed_children)
        self.completed_children = []
        self.failed = 0
        self.completed = 0

    def _policyOneFails(self, child):
        self.reset()
        return RESULT.FAIL

    def _policyAllFail(self, child):
        if self.failed >= len(self.children)+len(self.completed_children):
            self.reset()
            result = RESULT.FAIL
        else:
            result = RESULT.CONTINUE
        return result

    def _policyOneCompletes(self, child):
        self.reset()
        return RESULT.SUCCEED

    def _policyAllComplete(self, child):
        if self.completed >= len(self.children)+len(self.completed_children):
            self.reset()
            result = RESULT.SUCCEED
        else: 
            result = RESULT.CONTINUE
        return result

    def ifChildFails(self, child):
        self.failed += 1
        self.completed += 1
        self.children.remove(child)
        self.completed_children.append(child)
        return min(self.POLICY[self.failure_policy](child),
                   self.POLICY[self.completion_policy](child))

    def ifChildSucceeds(self, child):
        self.completed += 1
        self.children.remove(child)
        self.completed_children.append(child)
        return self.POLICY[self.completion_policy](child)

    def step(self, dt):
        results = []
        for c in tuple(self.children):
            c_result = c.step(dt)
            result = self.RESULT_TABLE[c_result](c)
            results.append(result)
        results.sort(reverse=True)
        return results[-1]


def logStdout(*stuff):
    print stuff

class Log(_BTDecorator):
    """Logging decorator.  Accepts a function w/ two arguments.

    Sends (self.child, result) to logFunc.
    """
    def __init__(self, child, logFunc=logStdout, **kwargs):
        super(Log, self).__init__(child, **kwargs)
        self.logFunc = logFunc

    def step(self, dt):
        result = self.child.step(dt)
        self.logFunc(self.child, result)

class _AfterNode(_BTNode):
    """Always return something until after so many steps.
    """
    def __init__(self, after=None, until_then=RESULT.CONTINUE, 
                 **kwargs):
        super(_AfterNode, self).__init__(**kwargs)
        self.until_then = until_then
        self.after = after
        self.count = 0

    def getFinalResult(self):
        return RESULT.CONTINUE

    def reset(self):
        count = 0

    def step(self, dt):
        if self.after is None:
            result = self.getFinalResult()
        else:
            if self.count >= self.after:
                self.reset()
                result = self.getFinalResult()
            else:
                result = self.until_then
            self.count += 1
        return result

class Succeed(_AfterNode):
    """Always succeed, or, succeed after a given number of steps.
    """
    def getFinalResult(self):
        return RESULT.SUCCEED

class Fail(_AfterNode):
    """Always fail, or, fail after a given number of steps.
    """
    def getFinalResult(self):
        return RESULT.FAIL


class Error(_AfterNode):
    """Always raise an exception, or, raise after a given number of steps.
    """
    def __init__(self, exc=Exception, msg="Reached Error node.", 
                 after=None, until_then=RESULT.CONTINUE, **kwargs):
        super(Error, self).__init__(after=after, until_then=until_then,
                                    **kwargs)
        self.exc = exc
        self.msg = msg

    def getFinalResult(self):
        raise self.exc(self.msg)

class Limit(_BTDecorator):
    def __init__(self, child, delay, **kwargs):
        super(Limit, self).__init__(child, **kwargs)
        self.delay = delay
        self.counter = 0.0 #delay*random.random()

    def step(self, dt):
        self.counter += dt
        delay = self.delay
        if self.counter >= delay:
            dt = self.counter
            self.counter -= delay
            result = self.child.step(dt)
        else:
            result = RESULT.CONTINUE
        return result


class FindNearest(_BTNode):
    def __init__(self, memory_point, categories=(), look_radius=400, **kwargs):
        super(FindNearest, self).__init__(**kwargs)
        self.look_radius = look_radius
        self.categories = categories
        self.memory_point = memory_point
    
    def step(self, dt):
        target = self.target
        x, y = target.xy

        found = []
        for c in self.categories:
            found.extend(self.blackboard[c])
        if found:
            nearby = collisions.collide_single((x, y, self.look_radius), found)
            if nearby:
                nearby.sort(key=lambda r: (abs(r.x - x)+abs(r.y -y)) / 2.0)
                self.memory[self.memory_point] = nearby[0].xy
                result = RESULT.SUCCEED
            else:
                self.memory[self.memory_point] = None
                result = RESULT.FAIL
        else:
            result = RESULT.FAIL
        return result

class _MemoryHandler(_BTNode):
    def __init__(self, memory_point, from_blackboard=False, **kwargs):
        super(_MemoryHandler, self).__init__(**kwargs)
        self.memory_point = memory_point

        if from_blackboard:
            self.getMemoryPoint = self._getBlackboardPoint
            self.setMemoryPoint = self._setBlackboardPoint

    def getMemoryPoint(self):
        return self.memory.setdefault(self.memory_point, None)

    def _getBlackboardPoint(self):
        return self.blackboard.setdefault(self.memory_point, None)

    def setMemoryPoint(self, v):
        self.memory[self.memory_point] = v

    def _setBlackboardPoint(self, v):
        self.blackboard[self.memory_point] = v



class CheckMemory(_MemoryHandler):
    """Check our memory to see if we know about the given memory point.
    """
    def step(self, dt):
        mp = self.getMemoryPoint()
        if mp is None:
            result = RESULT.FAIL
        else:
            result = RESULT.SUCCEED
        return result

class TurnToFace(_MemoryHandler):
    """Turn to face the position recalled from memory.
    """
    def step(self, dt):
        mp = self.getMemoryPoint()
        target = self.target
        if mp is None:
            result = RESULT.FAIL
        else:
            rot_delta = target.getRotationDelta(*mp)
                
            if rot_delta < -0.1:
                target.turning_right = False
                target.turning_left = True
            elif rot_delta > 0.1:
                target.turning_right = True
                target.turning_left = False
            else:
                target.turning_right = False
                target.turning_left = False

            result = RESULT.SUCCEED
        return result

class TurnToFlee(_MemoryHandler):
    """Turn to flee from the position recalled from memory.
    """
    def step(self, dt):
        mp = self.getMemoryPoint()
        target = self.target
        if mp is None:
            result = RESULT.FAIL
        else:
            rot_delta = target.getRotationDelta(*mp)
                
            if rot_delta < -0.1:
                target.turning_right = True
                target.turning_left = False
            elif rot_delta > 0.1:
                target.turning_right = False
                target.turning_left = True
            else:
                target.turning_right = False
                target.turning_left = False

            result = RESULT.SUCCEED
        return result

class StopRotation(_BTNode):
    def step(self, dt):
        target = self.target
        target.turning_right = False
        target.turning_left = False
        return RESULT.SUCCEED

class AmFacing(_MemoryHandler):
    def step(self, dt):
        mp = self.getMemoryPoint()
        target = self.target
        if mp is None:
            result = RESULT.FAIL
        else:
            rot_delta = target.getRotationDelta(*mp)

            if rot_delta > -0.1 and rot_delta < 0.1:
                result = RESULT.SUCCEED
            else:
                result = RESULT.FAIL
        return result


class AmFleeing(_MemoryHandler):
    def step(self, dt):
        mp = self.getMemoryPoint()
        target = self.target
        if mp is None:
            result = RESULT.FAIL
        else:
            rot_delta = abs(target.getRotationDelta(*mp))

            if rot_delta > (pi-0.1) and rot_delta < (pi+0.1):
                result = RESULT.SUCCEED
            else:
                result = RESULT.FAIL
        return result


class MoveForward(_BTNode):
    def __init__(self, **kwargs):
        super(MoveForward, self).__init__(**kwargs)

    def step(self, dt):
        target = self.target
        target.accelerating = True
        target.backing = False
        return RESULT.SUCCEED

class Stop(_BTNode):
    def __init__(self, **kwargs):
        super(Stop, self).__init__(**kwargs)

    def step(self, dt):
        target = self.target
        target.accelerating = False
        target.backing = False
        target.turning_left = False
        target.turning_right = False
        #self.report("Stopped.")
        return RESULT.SUCCEED


class Flip(_BTDecorator):
    """Flip the return value of the child.
    """
    def ifChildSucceeds(self, child):
        return RESULT.FAIL

    def ifChildFails(self, child):
        return RESULT.SUCCEED

class CheckCollision(_BTNode):
    def step(self, dt):
        target = self.target
        collide = self.target.last_collision
        if collide is not None:
            self.target.last_collision = None
            self.memory['last_collision'] = collide
            #self.report("last collided at %s" % (collide,))
            result = RESULT.SUCCEED
        else:
            #self.report("found no previous collisions.")
            result = RESULT.FAIL
        return result

class LookAheadForCollision(_BTNode):
    def __init__(self, look_ahead, **kwargs):
        super(LookAheadForCollision, self).__init__(**kwargs)
        self.look_ahead = look_ahead

    def step(self, dt):
        x,y = self.target.convert_offset((0, self.look_ahead))
        c = collisions.collide_single((x, y, 1), self.blackboard[EntityCategories.building])
        if c:
            result = RESULT.SUCCEED
        else:
            result = RESULT.FAIL
        return result
        

class CheckSimpleDistance(_MemoryHandler):
    """Checks for Manhattan distance.
    """
    def __init__(self, memory_point, 
                 greater_than=0.0, less_than=sys.maxint, **kwargs):
        super(CheckSimpleDistance, self).__init__(memory_point, **kwargs)
        self.greater_than = greater_than
        self.less_than = less_than
    
    def step(self, dt):
        my_x, my_y = self.target.xy
        mp = self.getMemoryPoint()
        if mp is not None:
            c_x, c_y = mp
        
            dx = abs(my_x - c_x)
            dy = abs(my_y - c_y)
            
            in_bounds = False
            gt = self.greater_than
            lt = self.less_than
            if gt < dx < lt or gt < dy < lt:
                in_bounds = True
            if in_bounds:
                result = RESULT.SUCCEED
            else:
                result = RESULT.FAIL
        else:
            result = RESULT.FAIL
        return result

class SetAttribute(_BTNode):
    def __init__(self, name, value_func, **kwargs):
        self.set_name = name
        self.getValue = value_func

    def step(self, dt):
        setattr(self.target, self.set_name, self.getValue())
        return RESULT.SUCCEED

class ClearMemory(_MemoryHandler):
    def step(self, dt):
        self.memory[self.memory_point] = None
        #self.report("cleared memory at %s" % (self.memory_point,))
        #self.report("remembers %s as %s" % (self.memory_point, 
        #                                    self.memory[self.memory_point]))
        return RESULT.SUCCEED

class TurnRandom(_BTNode):
    def step(self, dt):
        right = choice((True, False))
        target = self.target
        target.turning_right = right
        target.turning_left = not right
        return RESULT.SUCCEED


class factory(object):
    def __init__(self):
        msg = "Do not instantiate.  Use class members and methods."
        raise NotImplementedError, msg

    @classmethod
    def buildEnemyAI(cls, enemy):
        r = Root(enemy, 
                 Parallel(Limit(FindNearest('robot_pos',
                                            (EntityCategories.player,),
                                            look_radius=600),
                                0.3),
                        
                          Sequence(CheckMemory('robot_pos',),
                                   TurnToFace('robot_pos'),
                                   MoveForward(),
                                   AmFacing('robot_pos'),
                                   StopRotation(),
                                   SetAttribute('firing', lambda: True),
                                   ),
                          Sequence(Flip(CheckMemory('robot_pos')),
                                   Stop(),
                                   ),

                          Limit(FindNearest('dude_pos',
                                            (EntityCategories.other,)),
                                0.1),
                          Sequence(CheckMemory('dude_pos'),
                                   TurnToFace('dude_pos'),
                                   MoveForward(),
                                   AmFacing('dude_pos'),
                                   StopRotation(),
                                   ),

                          Sequence(CheckCollision(),
                                   Stop(),
                                   TurnToFlee('last_collision'),
                                   ClearMemory('last_collision'),
                                   MoveForward(),
                                   ),
                          Sequence(LookAheadForCollision(16),
                                   Stop(),
                                   TurnRandom(),
                                   MoveForward(),
                                   ),

                          )
                 )
        return r

    @classmethod
    def buildDudeAI(cls, dude):
        r = Root(dude, 
                 Limit(
                 Parallel(Limit(FindNearest('robot_pos', 
                                            (EntityCategories.player,
                                             EntityCategories.AI,)
                                            ), 0.1),
                          Sequence(CheckMemory('robot_pos'),
                                   TurnToFlee('robot_pos'),
                                   MoveForward(),
                                   AmFleeing('robot_pos'),
                                   StopRotation(),
                                   ),
                          Sequence(Flip(CheckMemory('robot_pos')),
                                   Stop(),
                                   ),
                          Sequence(CheckCollision(),
                                   Stop(),
                                   TurnToFlee('last_collision'),
                                   ClearMemory('last_collision'),
                                   MoveForward(),
                                   ),
                          Sequence(Limit(LookAheadForCollision(32), 0.2),
                                   TurnRandom(),
                                   ),

                          ),
                 random.uniform(0.1, 0.3)),
                 initial_delay=random.uniform(0.0, 0.8),
                 )
        # When we have found a player:
        # 
        return r
                 
