# -*- coding: utf-8 -*-
#    I, for one, welcome our new robot overloards
#    Copyright (C) 2008 David Eyk, David Thulson, Collin Sanford, David Howery
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""quadtreenode"""

import rabbyt.collisions
from rabbyt.primitives import Quad
from entity import EntityCategories

def splitQuadIntoQuarters(quad):
    left = quad.left
    right = quad.right
    top = quad.top
    bottom = quad.bottom
    x = quad.x
    # box.y appears to have a bug in Rabbyt 0.8.1
    # I posted here: http://groups.google.com/group/rabbyt/browse_thread/thread/ae2cfa3b97e87bdd
    # but I'm working around it for now... this seems to work
    y = (top + bottom) / 2.0

    return Quad((left, top, x, y)),\
        Quad((x, top, right, y)),\
        Quad((left, y, x, bottom)),\
        Quad((x, y, right, bottom))

def pointInQuad(point, quad):
    return ((point.x >= quad.left)
        and (point.x <= quad.right)
        and (point.y >= quad.bottom)
        and (point.y <= quad.top))

def quadsCollide(a, b):
    return len(rabbyt.collisions.aabb_collide_single(a, (b,))) != 0


class QuadtreeNode:
    def __init__(self, parent, bounding_box, size_threshold):
        self.bounding_box = bounding_box
        # Create a dictionary from each category to a value of 0
        # Can I do this in a more python-y way?
        self._category_counts = {}
        for c in EntityCategories:
            self._category_counts[c] = 0
        self._all_entities = None
        self._children = None
        self._parent = parent
        
        is_leaf = (bounding_box.width <= size_threshold) and (bounding_box.height <= size_threshold)
        if is_leaf:
            self._all_entities = []
        else:
            cb0, cb1, cb2, cb3 = splitQuadIntoQuarters(self.bounding_box)
            self._children = [
                QuadtreeNode(self, cb0, size_threshold),
                QuadtreeNode(self, cb1, size_threshold),
                QuadtreeNode(self, cb2, size_threshold),
                QuadtreeNode(self, cb3, size_threshold)
                ]
    
    def add(self, entity):
        if self._addHelper(entity):
            return self
        else:
            for child in self._children:
                if pointInQuad(entity, child.bounding_box):
                    return child.add(entity)
            assert False, "Internal error adding entity"
    
    def remove(self, entity):
        self._removeHelper(entity)
        if not self.isRoot():
            self._parent.remove(entity)
    
    def queryEntitiesNear(self, x, y, radius, category_set):
        q = Quad((x - radius, y + radius, x + radius, y - radius))
        result = self._queryWithinHelper(q, category_set)
        if result:
            s = (x, y, radius)
            result = rabbyt.collisions.collide_single(s, result)
        return result
    
    def queryEntitiesWithin(self, quad, category_set):
        result = self._queryWithinHelper(quad, category_set)
        if result:
            result = rabbyt.collisions.aabb_collide_single(quad, result)
        return result
    
    def updateEntity(self, entity):
        assert self.isLeaf(), "updateEntity should only be called directly on leaves"
        
        if pointInQuad(entity, self.bounding_box):
            return self
        else:
            self._removeHelper(entity)
            if self.isRoot():
                assert entity.category is not EntityCategories.player, "Player is no longer within world bounds"
                entity.kill()
                return None
            return self._parent._updateEntityRecurse(entity)
    
    def isRoot(self):
        return self._parent is None
    
    def isLeaf(self):
        return self._children is None
    
    def _entitiesInCategories(self, category_set):
        return [x for x in self._all_entities if x.category in category_set]
    
    def _queryWithinHelper(self, quad, category_set):
        result = []
        if quadsCollide(quad, self.bounding_box):
            if self.isLeaf():
                result.extend(self._entitiesInCategories(category_set))
            else:
                for child in self._children:
                    result.extend(child._queryWithinHelper(quad, category_set))
        return result
    
    def _removeHelper(self, entity):
        ent_cat = entity.category
        self._category_counts[ent_cat] = self._category_counts[ent_cat] - 1
        
        if self.isLeaf():
            self._all_entities.remove(entity)
            entity.setNode(None)
    
    def _addHelper(self, entity):
        if self.isRoot():
            assert pointInQuad(entity, self.bounding_box), "Entity is positioned outside of world boundaries"
        
        is_leaf = self.isLeaf()
        ent_cat = entity.category
        self._category_counts[ent_cat] = self._category_counts[ent_cat] + 1
        
        if is_leaf:
            self._all_entities.append(entity)
            entity.setNode(self)
        
        return is_leaf
    
    def _updateEntityRecurse(self, entity):
        if pointInQuad(entity, self.bounding_box):
            if self._addHelper(entity):
                return self
            else:
                for child in self._children:
                    if pointInQuad(entity, child.bounding_box):
                        return child.add(entity)
                assert False, "Internal error updating entity"
        else:
            self._removeHelper(entity)
            if self.isRoot():
                assert entity.category is not EntityCategories.player, "Player is no longer within world bounds"
                entity.kill()
                return None
            return self._parent._updateEntityRecurse(entity)
        
