# -*- coding: utf-8 -*-
#    I, for one, welcome our new robot overloards
#    Copyright (C) 2008 David Eyk, David Thulson, Collin Sanford, David Howery
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
""" A group of classes to display the terrain """

import re
import os
import shutil

from projectx import resources
from projectx.entity import Entity, EntityCategories
from projectx import rsprite
from projectx.sound import sound
import actors

def IncXY(x, y, max_x, max_y):
    rx, ry = x, y
    rx += 1
    if rx >= max_x:
        ry += 1
        rx = 0
    return rx, ry

class Tile(Entity):
    building_destroy_sound = sound.loadSound(resources.getFilePath('sfx/building-collapse.wav'))
    def __init__(self, 
                 texture,
                 damaged_texture,
                 blocked=False,
                 strength=0,
                 rotation=0,
                 location=(0,0),
                 category=EntityCategories.ground):
        category = EntityCategories.building if blocked else category
        super(Tile, self).__init__(texture=resources.loadImage(texture).texture,
                                   category=category)
        self._texture_name = os.path.basename(texture)
        self._damaged_name = os.path.basename(damaged_texture)
        self.damagedTexture = rsprite.RASprite(resources.loadImage(damaged_texture).texture)
        self.damagedTexture.alpha = 0.0
        self.damagedTexture.xy = location
        self.blocked = blocked
        self.strength = strength
        self.rot = rotation
        self.xy = location
        self.health = strength
        
    def collideWith(self, obj):
        if self.strength <> 0 and self.blocked and type(obj) is not actors.Robot:
            obj.damage(1)
        
    def damage(self, amt):
        if self.strength <> 0 and self.blocked:
            self.health -= amt
            if self.health <= 0:
                self.kill()
            self.damagedTexture.alpha = (self.strength - self.health) / float(self.strength)
    
    def render(self):
        Entity.render(self)
        self.damagedTexture.draw()
        
    def kill(self):
        Entity.kill(self)
        self.building_destroy_sound.play()
        self._world.change_points(20, 'Keep knocking buildings down!')

class Terrain(object):
    def __init__(self, World):
        self.__world = World
        self.__tiles = {}
        
    def _addTile(self, t):
        self.__world.add(t)
        #self.__tiles[t.xy] = t
    
    def LoadTerrain(self, config_text, fill_empty=True):
        r = re.compile(r'dim:(\d+)x(\d+)')
        s = r.match(config_text)
        self.dimensions = (int(s.group(1)), int(s.group(2)))
        
        x,y = 0,0
        g = re.findall('<(.+)>', config_text)
        for i in g:
            if i <> 'Blank':
                tile_list = i.split(',')
                t = Tile(tile_list[0].strip('" '),
                         tile_list[1].strip('" '),
                         bool(int(tile_list[2])),
                         int(tile_list[3]),
                         int(tile_list[4]),
                         (x * 64, y * 64))
                t.scale = (int(tile_list[5]), int(tile_list[6]))
                self._addTile(t)
            elif fill_empty:
                t = Tile('ground.png',
                         'ground.png',
                         False,
                         0,
                         0,
                         (x * 64, y * 64))
                self._addTile(t)
            x += 1
            if x >= self.dimensions[0]:
                x = 0
                y += 1
                if y >= self.dimensions[1]:
                    break
                
    def __sortTiles(self, x, y):
        if (x.x == y.x) and (x.y == y.y):
            return 0
        if x.y == y.y:
            return -1 if x.x < y.x else 1
        return -1 if x.y < y.y else 1

    def SaveTerrain(self, filename, tile_list):
        tile_list.sort(cmp=self.__sortTiles)
        l = tile_list
        max_x, min_x, max_y, min_y = 0,0,0,0
        for i in l:
            max_x = i.x if i.x > max_x else max_x
            min_x = i.x if i.x < min_x else min_x
            max_y = i.y if i.y > max_y else max_y
            min_y = i.y if i.y < min_y else min_y
            
        dimx = ((max_x - min_x) / 64) + 1
        dimy = ((max_y - min_y) / 64) + 1
            
        f = open(resources.getFilePath(filename + '_tmp'), 'w')
        try:
            x, y = 0, 0
            f.write('dim:%dx%d\n' % (dimx, dimy))
            for i in l:
                while (i.x <> x * 64) or (i.y <> y * 64):
                    f.write('<Blank>\n')
                    x, y = IncXY(x, y, dimx, dimy)
                    assert y + 1 < dimy, "You probably had a hidden tile someplace...sorry."
                    
                s = '<"%s","%s",%d,%d,%d,%d,%d>\n' % (
                    i._texture_name.strip('" '),
                    i._damaged_name.strip('" '),
                    #'destroyed.png',
                    int(i.blocked),
                    i.strength,
                    i.rot,
                    i.scale_x,
                    i.scale_y)
                f.write(s)
                x, y = IncXY(x, y, dimx, dimy)
        finally:
            f.close()
        
        os.remove(resources.getFilePath(filename))
        shutil.move(resources.getFilePath(filename + '_tmp'), resources.getFilePath(filename))
