# -*- coding: utf-8 -*-
#    I, for one, welcome our new robot overloards
#    Copyright (C) 2008 David Eyk, David Thulson, Collin Sanford, David Howery
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""world"""

import re

from quadtreenode import QuadtreeNode
from rabbyt.primitives import Quad
import rabbyt.collisions
from entity import EntityCategories
from entity import AllCategories
from entity import CollidableCategories
import terrain


def getOrderedPairByCategory(pair):
    if pair[0].category.index <= pair[1].category.index:
        return pair
    else:
        return (pair[1], pair[0])

# Rabbyt's collide method seems to be broken
# So this appears to be what it does from the docs
def myCollide(objects):
    if len(objects) <= 10:
        return rabbyt.collisions.brute_force(objects)
    else:
        result = []
        groups = rabbyt.collisions.rdc(objects, 0, 10)
        for g in groups:
            result.extend(rabbyt.collisions.brute_force(g))
        return result


class World:
    def __init__(self, filename):
        f = open(filename).read()
        r = re.compile(r'dim:(\d+)x(\d+)')
        s = r.match(f) 

        # Read params from a file?
        LEAF_SIZE = 1024
        WORLD_QUAD = Quad((0, (int(s.group(1)) + 1)*64, (int(s.group(2)) + 1)*64, 0))
        
        self._collide_dict = {}
        self._default_col_callback = None
        
        self._root = QuadtreeNode(None, WORLD_QUAD, LEAF_SIZE)
        # A dict from every category to an empty list for the value
        self._entity_dict = {}
        for c in EntityCategories:
            self._entity_dict[c] = []
            
        self._map = terrain.Terrain(self)
        self._map.LoadTerrain(f)
    
    def add(self, entity):
        self._entity_dict[entity.category].insert(0, entity)
        self._root.add(entity)
        entity.setWorld(self)
    
    def remove(self, entity):
        self._entity_dict[entity.category].remove(entity)
        entity._node.remove(entity)
        entity.setWorld(None)
    
    # This needs to be called once a frame
    def update(self):
        for l in self._entity_dict.values():
            for e in l:
                e.updateNode()
    
    def handleCollisions(self, category_set = CollidableCategories):
        candidates = self.queryAllEntities(category_set)
        pairs = myCollide(candidates)
        
        for p in pairs:
            ord_p = getOrderedPairByCategory(p)
            assert ord_p[0].category.index <= ord_p[1].category.index, "pair not ordered"
            a, b = ord_p
            callback = self._collide_dict.get((a.category, b.category), self._default_col_callback)
            if callback is not None:
                callback(a, b)
    
    def setDefaultCollisionHandler(self, callback):
        self._default_col_callback = callback
    
    def registerCollisionHandler(self, cat_a, cat_b, callback):
        if cat_a.index <= cat_b.index:
            key = (cat_a, cat_b)
        else:
            key = (cat_b, cat_a)
        self._collide_dict[key] = callback
    
    def queryAllEntities(self, category_set = AllCategories):
        # Is there a more python-y way to do this?
        result = []
        for c in category_set:
            result.extend(self._entity_dict[c])
        return result
    
    def queryEntitiesNear(self, x, y, radius, category_set = AllCategories):
        return self._root.queryEntitiesNear(x, y, radius, category_set)
    
    def queryEntitiesWithin(self, bound_box, category_set = AllCategories):
        return self._root.queryEntitiesWithin(bound_box, category_set)
