# -*- coding: utf-8 -*-
"""testquadtree
"""

import unittest
from projectx import quadtreenode
from rabbyt.primitives import Quad
from projectx import entity
from pyglet import image
from projectx import resources


def compareQuad(a, b):
    return ((a.left == b.left)
        and (a.right == b.right)
        and (a.top == b.top)
        and (a.bottom == b.bottom))

class QuadtreeNodeTest(unittest.TestCase):
    """Test QuadtreeNode
    """
    def setUp(self):
        self._image = car_img = image.load(resources.getFilePath('Street.png'))
        self._tex = self._image.texture
    
    def tearDown(self):
        pass
    
    # This should probably also be tested in the entity, but since it is needed
    #   by the quadtreenode, I am testing it here
    def testEntityCategories(self):
        """Test EntityCategories enum
        """
        self.assertEqual(len(entity.EntityCategories), 6)
        self.assertEqual(len(entity.EntityCategories), len(entity.AllCategories))
        s = set([ entity.EntityCategories.player, entity.EntityCategories.weapon ])
        inter = entity.AllCategories.intersection(s)
        self.assertEqual(len(inter), 2)
        s2 = set([ entity.EntityCategories.player, entity.EntityCategories.AI ])
        inter = s.intersection(s2)
        self.assertEqual(len(inter), 1)
        self.assert_(entity.EntityCategories.player in inter)

    def testSplitQuadIntoQuarters(self):
        """Test splitQuadIntoQuarters method
        """
        #self.assert_('b' in self.foo)
        #self.assertEqual(self.foo, 'bar')
        #self.assertNotEqual(math.pi, 3.1415)
        #self.assertAlmostEqual(math.pi, 3.1415, places=3)

        #raisesExc = lambda: len(5)
        #self.assertRaises(TypeError, raisesExc)
        q = Quad((0, 1, 1, 0))
        cNW, cNE, cSW, cSE = quadtreenode.splitQuadIntoQuarters(q)
        self.assert_(compareQuad(cNW, Quad((0.0, 1.0, 0.5, 0.5))))
        self.assert_(compareQuad(cNE, Quad((0.5, 1.0, 1.0, 0.5))))
        self.assert_(compareQuad(cSW, Quad((0.0, 0.5, 0.5, 0.0))))
        self.assert_(compareQuad(cSE, Quad((0.5, 0.5, 1.0, 0.0))))
        
        q = Quad((0, -1, -1, 0))
        cNW, cNE, cSW, cSE = quadtreenode.splitQuadIntoQuarters(q)
        self.assert_(compareQuad(cNW, Quad((-1.0, 0.0, -0.5, -0.5))))
        self.assert_(compareQuad(cNE, Quad((-0.5, 0.0, 0.0, -0.5))))
        self.assert_(compareQuad(cSW, Quad((-1.0, -0.5, -0.5, -1.0))))
        self.assert_(compareQuad(cSE, Quad((-0.5, -0.5, 0.0, -1.0))))
    
    def testPointInQuad(self):
        class PointTest(object):
            def __init__(self, x, y):
                self.x = x
                self.y = y
        
        q = Quad((0, 10, 10, 0))
        p = PointTest(-1, -1)
        self.assert_(not quadtreenode.pointInQuad(p, q))
        
        p.x = 5
        p.y = 5
        self.assert_(quadtreenode.pointInQuad(p, q))
        
        p.y = 15
        self.assert_(not quadtreenode.pointInQuad(p, q))
    
    def testQuadsCollide(self):
        a = Quad((0, 10, 10, 0))
        b = Quad((5, 15, 15, 5))
        c = Quad((50, 60, 60, 50))
        
        self.assert_(quadtreenode.quadsCollide(a, b))
        self.assert_(not quadtreenode.quadsCollide(a, c))
    
    def testQuadtreeCreation(self):
        q = Quad((0, 600, 600, 0))
        root = quadtreenode.QuadtreeNode(None, q, 1000)
        self.assert_(root.isRoot())
        self.assert_(root.isLeaf())
        
        root = quadtreenode.QuadtreeNode(None, q, 400)
        self.assert_(root.isRoot())
        self.assert_(not root.isLeaf())
        for c in root._children:
            self.assert_(not c.isRoot())
            self.assert_(c._parent is root)
            self.assert_(c.isLeaf())
        
        root = quadtreenode.QuadtreeNode(None, q, 200)
        self.assert_(root.isRoot())
        self.assert_(not root.isLeaf())
        for c in root._children:
            self.assert_(not c.isRoot())
            self.assert_(c._parent is root)
            self.assert_(not c.isLeaf())
            
            for gc in c._children:
                self.assert_(not gc.isRoot())
                self.assert_(gc._parent is c)
                self.assert_(gc.isLeaf())
    
    def testQuadtreeAddRemove(self):
        q = Quad((0, 600, 600, 0))
        root = quadtreenode.QuadtreeNode(None, q, 200)
        e = entity.Entity(self._tex, entity.EntityCategories.player)
        e.x = 599
        e.y = 599
        root.add(e)
        self.assert_(root.isRoot())
        self.assert_(not root.isLeaf())
        self.assert_(root._all_entities is None)
        count = 0
        for c in root._children:
            self.assert_(not c.isRoot())
            self.assert_(not c.isLeaf())
            self.assert_(c._all_entities is None)
            for gc in c._children:
                self.assert_(not gc.isRoot())
                self.assert_(gc.isLeaf())
                self.assert_(gc._all_entities is not None)
                l = len(gc._all_entities)
                count = count + l
                if l != 0:
                    self.assert_(gc._all_entities[0] is e)
                    self.assert_(quadtreenode.pointInQuad(e, gc.bounding_box))
                    self.assert_(e._node is gc)
        self.assertEqual(count, 1)
        self.assert_(e._node is not None)
        
        e._node.remove(e)
        count = 0
        for c in root._children:
            self.assert_(not c.isRoot())
            self.assert_(not c.isLeaf())
            self.assert_(c._all_entities is None)
            for gc in c._children:
                self.assert_(not gc.isRoot())
                self.assert_(gc.isLeaf())
                self.assert_(gc._all_entities is not None)
                count = count + len(gc._all_entities)
        self.assertEqual(count, 0)
        self.assert_(e._node is None)
    
    def testCandidatesWithin(self):
        q = Quad((0, 400, 400, 0))
        root = quadtreenode.QuadtreeNode(None, q, 100)
        self.assert_(root.isRoot())
        self.assert_(not root.isLeaf())
        for c in root._children:
            self.assert_(not c.isRoot())
            self.assert_(c._parent is root)
            self.assert_(not c.isLeaf())
            for gc in c._children:
                self.assert_(not gc.isRoot())
                self.assert_(gc._parent is c)
                self.assert_(gc.isLeaf())
        eNW = entity.Entity(self._tex, entity.EntityCategories.player)
        eNW.xy = (120, 280)
        root.add(eNW)
        
        eNE = entity.Entity(self._tex, entity.EntityCategories.AI)
        eNE.xy = (280, 280)
        root.add(eNE)
        
        eSW = entity.Entity(self._tex, entity.EntityCategories.weapon)
        eSW.xy = (120, 120)
        root.add(eSW)
        
        eSE = entity.Entity(self._tex, entity.EntityCategories.other)
        eSE.xy = (280, 120)
        root.add(eSE)
        
        count = 0
        for c in root._children:
            self.assert_(not c.isRoot())
            self.assert_(not c.isLeaf())
            self.assert_(c._all_entities is None)
            for gc in c._children:
                self.assert_(not gc.isRoot())
                self.assert_(gc.isLeaf())
                self.assert_(gc._all_entities is not None)
                l = len(gc._all_entities)
                # Make sure each entity is in it's own node
                self.assert_(l <= 1)
                count = count + l
        self.assertEqual(count, 4)
        
        q = Quad((190, 210, 210, 190))
        candidates = root._queryWithinHelper(q, entity.AllCategories)
        self.assertEqual(len(candidates), 4)
        candidates = root._queryWithinHelper(q, set((entity.EntityCategories.player,)))
        self.assertEqual(len(candidates), 1)
        candidates = root._queryWithinHelper(
            q,
            set((entity.EntityCategories.player, entity.EntityCategories.AI)))
        self.assertEqual(len(candidates), 2)
        
        q = Quad((240, 320, 320, 240))
        candidates = root._queryWithinHelper(q, entity.AllCategories)
        self.assertEqual(len(candidates), 1)
        self.assert_(candidates[0] is eNE)

    def testQueryEntitiesNear(self):
        q = Quad((0, 400, 400, 0))
        root = quadtreenode.QuadtreeNode(None, q, 100)
        self.assert_(root.isRoot())
        self.assert_(not root.isLeaf())
        for c in root._children:
            self.assert_(not c.isRoot())
            self.assert_(c._parent is root)
            self.assert_(not c.isLeaf())
            for gc in c._children:
                self.assert_(not gc.isRoot())
                self.assert_(gc._parent is c)
                self.assert_(gc.isLeaf())
        eNW = entity.Entity(self._tex, entity.EntityCategories.player)
        eNW.xy = (120, 280)
        root.add(eNW)
        
        eNE = entity.Entity(self._tex, entity.EntityCategories.AI)
        eNE.xy = (280, 280)
        root.add(eNE)
        
        eSW = entity.Entity(self._tex, entity.EntityCategories.weapon)
        eSW.xy = (120, 120)
        root.add(eSW)
        
        eSE = entity.Entity(self._tex, entity.EntityCategories.other)
        eSE.xy = (280, 120)
        root.add(eSE)
        
        count = 0
        for c in root._children:
            self.assert_(not c.isRoot())
            self.assert_(not c.isLeaf())
            self.assert_(c._all_entities is None)
            for gc in c._children:
                self.assert_(not gc.isRoot())
                self.assert_(gc.isLeaf())
                self.assert_(gc._all_entities is not None)
                l = len(gc._all_entities)
                # Make sure each entity is in it's own node
                self.assert_(l <= 1)
                count = count + l
        self.assertEqual(count, 4)
        
        ents = root.queryEntitiesNear(200, 200, 20, entity.AllCategories)
        self.assertEqual(len(ents), 0)
        
        ents = root.queryEntitiesNear(280, 280, 80, entity.AllCategories)
        self.assertEqual(len(ents), 1)
        self.assert_(ents[0] is eNE)
    
    def testQueryEntitiesWithin(self):
        q = Quad((0, 400, 400, 0))
        root = quadtreenode.QuadtreeNode(None, q, 100)
        self.assert_(root.isRoot())
        self.assert_(not root.isLeaf())
        for c in root._children:
            self.assert_(not c.isRoot())
            self.assert_(c._parent is root)
            self.assert_(not c.isLeaf())
            for gc in c._children:
                self.assert_(not gc.isRoot())
                self.assert_(gc._parent is c)
                self.assert_(gc.isLeaf())
        eNW = entity.Entity(self._tex, entity.EntityCategories.player)
        eNW.xy = (120, 280)
        root.add(eNW)
        
        eNE = entity.Entity(self._tex, entity.EntityCategories.AI)
        eNE.xy = (280, 280)
        root.add(eNE)
        
        eSW = entity.Entity(self._tex, entity.EntityCategories.weapon)
        eSW.xy = (120, 120)
        root.add(eSW)
        
        eSE = entity.Entity(self._tex, entity.EntityCategories.other)
        eSE.xy = (280, 120)
        root.add(eSE)
        
        count = 0
        for c in root._children:
            self.assert_(not c.isRoot())
            self.assert_(not c.isLeaf())
            self.assert_(c._all_entities is None)
            for gc in c._children:
                self.assert_(not gc.isRoot())
                self.assert_(gc.isLeaf())
                self.assert_(gc._all_entities is not None)
                l = len(gc._all_entities)
                # Make sure each entity is in it's own node
                self.assert_(l <= 1)
                count = count + l
        self.assertEqual(count, 4)
        
        q = Quad((190, 210, 210, 190))
        ents = root.queryEntitiesWithin(q, entity.AllCategories)
        self.assertEqual(len(ents), 0)
        
        q = Quad((240, 320, 320, 240))
        ents = root.queryEntitiesWithin(q, entity.AllCategories)
        self.assertEqual(len(ents), 1)
        self.assert_(ents[0] is eNE)
    
    def testUpdateEntity(self):
        q = Quad((0, 600, 600, 0))
        root = quadtreenode.QuadtreeNode(None, q, 200)
        e = entity.Entity(self._tex, entity.EntityCategories.player)
        e.x = 599
        e.y = 599
        root.add(e)
        self.assert_(root.isRoot())
        self.assert_(not root.isLeaf())
        self.assert_(root._all_entities is None)
        count = 0
        orig_node = None
        for c in root._children:
            self.assert_(not c.isRoot())
            self.assert_(not c.isLeaf())
            self.assert_(c._all_entities is None)
            for gc in c._children:
                self.assert_(not gc.isRoot())
                self.assert_(gc.isLeaf())
                self.assert_(gc._all_entities is not None)
                l = len(gc._all_entities)
                count = count + l
                if l != 0:
                    self.assert_(gc._all_entities[0] is e)
                    self.assert_(quadtreenode.pointInQuad(e, gc.bounding_box))
                    self.assert_(e._node is gc)
                    orig_node = gc
        self.assertEqual(count, 1)
        self.assert_(e._node is not None)
        self.assert_(orig_node is not None)
        self.assert_(e._node is orig_node)
        self.assert_(orig_node.isLeaf())
        
        orig_node.updateEntity(e)
        self.assert_(e._node is orig_node)
        self.assert_(orig_node._all_entities[0] is e)
        
        count = 0
        for c in root._children:
            self.assert_(not c.isRoot())
            self.assert_(not c.isLeaf())
            self.assert_(c._all_entities is None)
            for gc in c._children:
                self.assert_(not gc.isRoot())
                self.assert_(gc.isLeaf())
                self.assert_(gc._all_entities is not None)
                l = len(gc._all_entities)
                count = count + l
                if l != 0:
                    self.assert_(gc._all_entities[0] is e)
                    self.assert_(quadtreenode.pointInQuad(e, gc.bounding_box))
                    self.assert_(e._node is gc)
        self.assertEqual(count, 1)
        
        e.x = 1
        e.y = 1
        new_node = orig_node.updateEntity(e)
        self.assert_(new_node is not None)
        self.assert_(new_node is not orig_node)
        self.assert_(new_node.isLeaf())
        self.assert_(e._node is new_node)
        self.assertEqual(len(orig_node._all_entities), 0)

if __name__ == "__main__":
    # It's nice to be able to just run this test module by itself.
    # I like testoob, because it colorizes.
    import testoob
    testoob.main()
